/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2000 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "EmRegsVZTemp.h"
#include "EmRegsVZPrv.h"

#include "EmSPISlaveADS784x.h"	// EmSPISlaveADS784x

const int		kNumButtonRows = 3;
const int		kNumButtonCols = 4;


#pragma mark -

#define hwrVZGenericMonoPortBALARM_LED	0x40	// (L) Alarm LED

#define	hwrVZGenericMonoPortDSpi1Card	0x20	// (-) SPI #1, slot 0, SD/MMC Card Detect EOR Wake EOR Ready (aka IRQ2)
#define	hwrVZGenericMonoPortDPowerFail	0x80	// (L) Power Fail interrupt	(aka IRQ6)

#define	hwrVZGenericMonoPortGEL_ON		0x02	// (L) EL_ON
#define	hwrVZGenericMonoPortGADC_CS_N	0x04	// (H) ADC_CS#
#define	hwrVZGenericMonoPortG232_SHDN_N	0x08	// (L) 232_SHDN#

#define	hwrVZGenericMonoPortKLCDOn		0x02	// (-) LCD_DISP_ON (RW#)
#define hwrVZGenericMonoPortKAC_PWR_N	0x04	// (-) AC_PWR# (LDS#)
#define	hwrVZGenericMonoPortKVibeEn		0x10	// (-) VIBE_EN
#define	hwrVZGenericMonoPortKKbdRow0	0x20	// (H) Keyboard Row 0
#define	hwrVZGenericMonoPortKKbdRow1	0x40	// (H) Keyboard Row 1
#define	hwrVZGenericMonoPortKKbdRow2	0x80	// (H) Keyboard Row 2


const uint16	kGenericMonoMap[kNumButtonRows][kNumButtonCols] =
{
	{ keyBitHard1,	keyBitHard2,	keyBitHard3,	keyBitHard4 },
	{ keyBitPageUp,	keyBitPageDown,	0,				0 },
	{ keyBitPower,	keyBitContrast,	0,				0 }
};


// ---------------------------------------------------------------------------
//		 EmRegsVZGenericMono::EmRegsVZGenericMono
// ---------------------------------------------------------------------------

EmRegsVZGenericMono::EmRegsVZGenericMono (void) :
	EmRegsVZ (),
	fSPISlaveADC (new EmSPISlaveADS784x (kChannelSet1))
{
}


// ---------------------------------------------------------------------------
//		 EmRegsVZGenericMono::~EmRegsVZGenericMono
// ---------------------------------------------------------------------------

EmRegsVZGenericMono::~EmRegsVZGenericMono (void)
{
	delete fSPISlaveADC;
}


// ---------------------------------------------------------------------------
//		 EmRegsVZGenericMono::GetLCDScreenOn
// ---------------------------------------------------------------------------

Bool EmRegsVZGenericMono::GetLCDScreenOn (void)
{
	return (READ_REGISTER (portKData) & hwrVZGenericMonoPortKLCDOn) != 0;
}


// ---------------------------------------------------------------------------
//		 EmRegsVZGenericMono::GetLCDBacklightOn
// ---------------------------------------------------------------------------

Bool EmRegsVZGenericMono::GetLCDBacklightOn (void)
{
	return (READ_REGISTER (portGData) & hwrVZGenericMonoPortGEL_ON) != 0;
}


// ---------------------------------------------------------------------------
//		 EmRegsVZGenericMono::GetSerialPortOn
// ---------------------------------------------------------------------------

Bool EmRegsVZGenericMono::GetSerialPortOn (int uartNum)
{
	/*
		From Michael Cortopassi:

			UART 1 for IRda
			UART 2 for hotsync
	*/

	if (uartNum == 1)
		return (READ_REGISTER (portGData) & hwrVZGenericMonoPortG232_SHDN_N) != 0;

	return false;
}


// ---------------------------------------------------------------------------
//		 EmRegsVZGenericMono::GetVibrateOn
// ---------------------------------------------------------------------------

Bool EmRegsVZGenericMono::GetVibrateOn (void)
{
	return (READ_REGISTER (portKData) & hwrVZGenericMonoPortKVibeEn) != 0;
}


// ---------------------------------------------------------------------------
//		 EmRegsVZGenericMono::GetLEDState
// ---------------------------------------------------------------------------

uint16 EmRegsVZGenericMono::GetLEDState (void)
{
	uint16	result		= kLEDOff;
	UInt8	portBData	= READ_REGISTER (portBData);

	if (portBData & hwrVZGenericMonoPortBALARM_LED)
		result |= kLEDRed;

	return result;
}


// ---------------------------------------------------------------------------
//		 EmRegsVZGenericMono::GetPortInputValue
// ---------------------------------------------------------------------------
// Return the GPIO values for the pins on the port.  These values are used
// if the select pins are high.

uint8 EmRegsVZGenericMono::GetPortInputValue (int port)
{
	uint8	result = EmRegsVZ::GetPortInputValue (port);

	if (port == 'K')
	{
		// Make sure hwrVZPortKAC_PWR_N is set, or the dock-status routines
		// will report that we're powered (even if we aren't docked!).

		result |= hwrVZGenericMonoPortKAC_PWR_N;
	}

	return result;
}


// ---------------------------------------------------------------------------
//		 EmRegsVZGenericMono::GetPortInternalValue
// ---------------------------------------------------------------------------
// Return the dedicated values for the pins on the port.  These values are
// used if the select pins are low.

uint8 EmRegsVZGenericMono::GetPortInternalValue (int port)
{
	uint8	result = EmRegsVZ::GetPortInternalValue (port);

	if (port == 'D')
	{
		// Make sure that hwrVZGenericMonoPortDPowerFail is set.  If it's clear,
		// the battery code will make the device go to sleep immediately.
		//
		// Also make sure that hwrVZGenericMonoPortDSpi1Card is set.  If it's clear,
		// the slot driver will think there's a card installed and will try querying it.

		result |= hwrVZGenericMonoPortDPowerFail | hwrVZGenericMonoPortDSpi1Card;
	}

	return result;
}


// ---------------------------------------------------------------------------
//		 EmRegsVZGenericMono::GetKeyInfo
// ---------------------------------------------------------------------------

void EmRegsVZGenericMono::GetKeyInfo (int* numRows, int* numCols,
								uint16* keyMap, Bool* rows)
{
	*numRows = kNumButtonRows;
	*numCols = kNumButtonCols;

	memcpy (keyMap, kGenericMonoMap, sizeof (kGenericMonoMap));

	// Determine what row is being asked for.

	UInt8	portKDir	= READ_REGISTER (portKDir);
	UInt8	portKData	= READ_REGISTER (portKData);

	rows[0]	= (portKDir & hwrVZGenericMonoPortKKbdRow0) != 0 && (portKData & hwrVZGenericMonoPortKKbdRow0) == 0;
	rows[1]	= (portKDir & hwrVZGenericMonoPortKKbdRow1) != 0 && (portKData & hwrVZGenericMonoPortKKbdRow1) == 0;
	rows[2]	= (portKDir & hwrVZGenericMonoPortKKbdRow2) != 0 && (portKData & hwrVZGenericMonoPortKKbdRow2) == 0;
}


// ---------------------------------------------------------------------------
//		 EmRegsVZGenericMono::GetSPISlave
// ---------------------------------------------------------------------------

EmSPISlave* EmRegsVZGenericMono::GetSPISlave (void)
{
	if ((READ_REGISTER (portGData) & hwrVZGenericMonoPortGADC_CS_N) == 0)
	{
		return fSPISlaveADC;
	}

	return NULL;
}


#pragma mark -

// ---------------------------------------------------------------------------
//		 EmRegsVZGenericColor::GetLCDScreenOn
// ---------------------------------------------------------------------------

Bool EmRegsVZGenericColor::GetLCDScreenOn (void)
{
	// Override the Dragonball version and let the SED 1376 handle it.

	return EmHALHandler::GetLCDScreenOn ();
}


// ---------------------------------------------------------------------------
//		 EmRegsVZGenericColor::GetLCDBacklightOn
// ---------------------------------------------------------------------------

Bool EmRegsVZGenericColor::GetLCDBacklightOn (void)
{
	// Override the Dragonball version and let the SED 1376 handle it.

	return EmHALHandler::GetLCDBacklightOn ();
}


// ---------------------------------------------------------------------------
//		 EmRegsVZGenericColor::GetLCDHasFrame
// ---------------------------------------------------------------------------

Bool EmRegsVZGenericColor::GetLCDHasFrame (void)
{
	// Override the Dragonball version and let the SED 1376 handle it.

	return EmHALHandler::GetLCDHasFrame ();
}


// ---------------------------------------------------------------------------
//		 EmRegsVZGenericColor::GetLCDBeginEnd
// ---------------------------------------------------------------------------

void EmRegsVZGenericColor::GetLCDBeginEnd (emuptr& begin, emuptr& end)
{
	// Override the Dragonball version and let the SED 1376 handle it.

	EmHALHandler::GetLCDBeginEnd (begin, end);
}


// ---------------------------------------------------------------------------
//		 EmRegsVZGenericColor::GetLCDScanlines
// ---------------------------------------------------------------------------

void EmRegsVZGenericColor::GetLCDScanlines (EmScreenUpdateInfo& info)
{
	// Override the Dragonball version and let the SED 1376 handle it.

	EmHALHandler::GetLCDScanlines (info);
}



#pragma mark -

// Given a register (specified by its field name), return its address
// in emulated space.

#undef addressof
#define addressof(reg)				\
	(this->GetAddressStart () + fRegs.offsetof_##reg ())


// Macro to help the installation of handlers for a register.

#undef INSTALL_HANDLER
#define INSTALL_HANDLER(read, write, reg)			\
	this->SetHandler (	(ReadFunction) &EmRegsUSBPhilipsPDIUSBD12::read,		\
						(WriteFunction) &EmRegsUSBPhilipsPDIUSBD12::write,		\
						addressof (reg),			\
						fRegs.reg.GetSize ())


// ---------------------------------------------------------------------------
//		 EmRegsUSBPhilipsPDIUSBD12::EmRegsUSBPhilipsPDIUSBD12
// ---------------------------------------------------------------------------

EmRegsUSBPhilipsPDIUSBD12::EmRegsUSBPhilipsPDIUSBD12 (emuptr baseAddr) :
	fBaseAddr (baseAddr)
{
}


// ---------------------------------------------------------------------------
//		 EmRegsUSBPhilipsPDIUSBD12::~EmRegsUSBPhilipsPDIUSBD12
// ---------------------------------------------------------------------------

EmRegsUSBPhilipsPDIUSBD12::~EmRegsUSBPhilipsPDIUSBD12 (void)
{
}


// ---------------------------------------------------------------------------
//		 EmRegsUSBPhilipsPDIUSBD12::Initialize
// ---------------------------------------------------------------------------

void EmRegsUSBPhilipsPDIUSBD12::Initialize (void)
{
	EmRegs::Initialize ();
}


// ---------------------------------------------------------------------------
//		 EmRegsUSBPhilipsPDIUSBD12::Reset
// ---------------------------------------------------------------------------

void EmRegsUSBPhilipsPDIUSBD12::Reset (Bool hardwareReset)
{
	EmRegs::Reset (hardwareReset);

	if (hardwareReset)
	{
		memset (fRegs.GetPtr (), 0, fRegs.GetSize ());
	}
}


// ---------------------------------------------------------------------------
//		 EmRegsUSBPhilipsPDIUSBD12::Save
// ---------------------------------------------------------------------------

void EmRegsUSBPhilipsPDIUSBD12::Save (SessionFile& f)
{
	EmRegs::Save (f);
}


// ---------------------------------------------------------------------------
//		 EmRegsUSBPhilipsPDIUSBD12::Load
// ---------------------------------------------------------------------------

void EmRegsUSBPhilipsPDIUSBD12::Load (SessionFile& f)
{
	EmRegs::Load (f);
}


// ---------------------------------------------------------------------------
//		 EmRegsUSBPhilipsPDIUSBD12::Dispose
// ---------------------------------------------------------------------------

void EmRegsUSBPhilipsPDIUSBD12::Dispose (void)
{
	EmRegs::Dispose ();
}


// ---------------------------------------------------------------------------
//		 EmRegsUSBPhilipsPDIUSBD12::SetSubBankHandlers
// ---------------------------------------------------------------------------

void EmRegsUSBPhilipsPDIUSBD12::SetSubBankHandlers (void)
{
	// Install base handlers.

	EmRegs::SetSubBankHandlers ();

	// Now add standard/specialized handers for the defined registers.

	INSTALL_HANDLER (StdReadBE,			StdWriteBE,				data);
	INSTALL_HANDLER (StdReadBE,			StdWriteBE,				command);
}


// ---------------------------------------------------------------------------
//		 EmRegsUSBPhilipsPDIUSBD12::GetRealAddress
// ---------------------------------------------------------------------------

uint8* EmRegsUSBPhilipsPDIUSBD12::GetRealAddress (emuptr address)
{
	return (uint8*) fRegs.GetPtr () + address - this->GetAddressStart ();
}


// ---------------------------------------------------------------------------
//		 EmRegsUSBPhilipsPDIUSBD12::GetAddressStart
// ---------------------------------------------------------------------------

emuptr EmRegsUSBPhilipsPDIUSBD12::GetAddressStart (void)
{
	return fBaseAddr;
}


// ---------------------------------------------------------------------------
//		 EmRegsUSBPhilipsPDIUSBD12::GetAddressRange
// ---------------------------------------------------------------------------

uint32 EmRegsUSBPhilipsPDIUSBD12::GetAddressRange (void)
{
	return fRegs.GetSize ();
}

