/* -*- mode: C++; tab-width: 4 -*- */
/* ===================================================================== *\
	Copyright (c) 1998-2000 Palm, Inc. or its subsidiaries.
	All rights reserved.

	This file is part of the Palm OS Emulator.

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.
\* ===================================================================== */

#include "EmCommon.h"
#include "EmDlgFltk.h"

#include "Miscellaneous.h"		// ReplaceString, FormatString
#include "Platform.h"			// Platform::GetString
#include "Strings.r.h"			// kStr_
#include <FL/Enumerations.H>
#include <FL/Fl.H>
#include <FL/Fl_Adjuster.H>
#include <FL/Fl_Bitmap.H>
#include <FL/Fl_Box.H>
#include <FL/Fl_Browser.H>
#include <FL/Fl_Browser_.H>
#include <FL/Fl_Button.H>
#include <FL/Fl_Chart.H>
#include <FL/Fl_Check_Button.H>
#include <FL/Fl_Choice.H>
#include <FL/Fl_Clock.H>
#include <FL/Fl_Color_Chooser.H>
#include <FL/Fl_Counter.H>
#include <FL/Fl_Dial.H>
#include <FL/Fl_Double_Window.H>
#include <FL/Fl_Fill_Dial.H>
#include <FL/Fl_Fill_Slider.H>
#include <FL/Fl_Float_Input.H>
#include <FL/Fl_FormsBitmap.H>
#include <FL/Fl_FormsPixmap.H>
#include <FL/Fl_Free.H>
#include <FL/Fl_Gl_Window.H>
#include <FL/Fl_Group.H>
#include <FL/Fl_Hold_Browser.H>
#include <FL/Fl_Hor_Fill_Slider.H>
#include <FL/Fl_Hor_Nice_Slider.H>
#include <FL/Fl_Hor_Slider.H>
#include <FL/Fl_Hor_Value_Slider.H>
#include <FL/Fl_Image.H>
#include <FL/Fl_Input.H>
#include <FL/Fl_Input_.H>
#include <FL/Fl_Int_Input.H>
#include <FL/Fl_Light_Button.H>
#include <FL/Fl_Line_Dial.H>
#include <FL/Fl_Menu.H>
#include <FL/Fl_Menu_.H>
#include <FL/Fl_Menu_Bar.H>
#include <FL/Fl_Menu_Button.H>
#include <FL/Fl_Menu_Item.H>
#include <FL/Fl_Menu_Window.H>
#include <FL/Fl_Multi_Browser.H>
#include <FL/Fl_Multi_Label.H>
#include <FL/Fl_Multiline_Input.H>
#include <FL/Fl_Multiline_Output.H>
#include <FL/Fl_Nice_Slider.H>
#include <FL/Fl_Object.H>
#include <FL/Fl_Output.H>
#include <FL/Fl_Overlay_Window.H>
#include <FL/Fl_Pack.H>
#include <FL/Fl_Pixmap.H>
#include <FL/Fl_Positioner.H>
#include <FL/Fl_Radio_Button.H>
#include <FL/Fl_Radio_Light_Button.H>
#include <FL/Fl_Radio_Round_Button.H>
#include <FL/Fl_Repeat_Button.H>
#include <FL/Fl_Return_Button.H>
#include <FL/Fl_Roller.H>
#include <FL/Fl_Round_Button.H>
#include <FL/Fl_Round_Clock.H>
#include <FL/Fl_Scroll.H>
#include <FL/Fl_Scrollbar.H>
#include <FL/Fl_Secret_Input.H>
#include <FL/Fl_Select_Browser.H>
#include <FL/Fl_Simple_Counter.H>
#include <FL/Fl_Single_Window.H>
#include <FL/Fl_Slider.H>
#include <FL/Fl_Tabs.H>
#include <FL/Fl_Tile.H>
#include <FL/Fl_Timer.H>
#include <FL/Fl_Toggle_Button.H>
#include <FL/Fl_Toggle_Light_Button.H>
#include <FL/Fl_Toggle_Round_Button.H>
#include <FL/Fl_Valuator.H>
#include <FL/Fl_Value_Input.H>
#include <FL/Fl_Value_Output.H>
#include <FL/Fl_Value_Slider.H>
#include <FL/Fl_Widget.H>
#include <FL/Fl_Window.H>
#include <FL/fl_ask.H>
#include <FL/fl_draw.H>
#include <FL/fl_file_chooser.H>
#include <FL/fl_message.H>
#include <FL/fl_show_colormap.H>
#include <FL/fl_show_input.H>

#include "EmDlgFltkFactory.h"

#include "espws-2.0/FileChooser.h"

#include <list>
#include <map>
#include <string>
#include <utility>		// pair
#include <queue>

typedef string			FilterList;

static FilterList		PrvConvertTypeList (const EmFileTypeList&);
static void				PrvAddFilter (FilterList& filter, const char* pattern);
static Fl_Window*		PrvMakeDialog (EmDlgID);
static void				PrvIdleCallback (void* data);

// -----------------------------------------------------------------------------
// helper routines for establishing z-order independent ids for widgets.

typedef map<EmDlgItemID, Fl_Widget*> ID2WidgetType;
typedef map<Fl_Widget*, EmDlgItemID> Widget2IDType;

ID2WidgetType gID2Widget;
Widget2IDType gWidget2ID;

Fl_Widget* PrvFindWidgetByID (EmDlgItemID id)
{
	ID2WidgetType::iterator iter = gID2Widget.find (id);
	if (iter == gID2Widget.end ())
		return NULL;

	return iter->second;
}

EmDlgItemID PrvFindIDByWidget (Fl_Widget* o)
{
	Widget2IDType::iterator iter = gWidget2ID.find (o);
	if (iter == gWidget2ID.end ())
		return kDlgItemNone;

	return iter->second;
}

void PrvSetWidgetID (Fl_Widget* o, EmDlgItemID id)
{
	gID2Widget[id] = o;
	gWidget2ID[o] = id;
}

void PrvClearWidgetID (EmDlgItemID id)
{
	Fl_Widget* o = ::PrvFindWidgetByID (id);

	if (o != NULL)
	{
		gID2Widget.erase (id);
		gWidget2ID.erase (o);
	}
}

void PrvClearWidgetID (Fl_Widget* o)
{
	EmDlgItemID id = ::PrvFindIDByWidget (o);

	if (id != kDlgItemNone)
	{
		gID2Widget.erase (id);
		gWidget2ID.erase (o);
	}
}

void PrvClearAllWidgetIDs (void)
{
	gID2Widget.clear ();
	gWidget2ID.clear ();
}

void PrvClearDlgWidgetIDs (Fl_Group* w)
{
	// Clear self.
	::PrvClearWidgetID (w);

	// Clear children.
	for (int ii = 0; ii < w->children (); ++ii)
	{
		Fl_Widget* child = w->child (ii);

		Fl_Group* g = dynamic_cast<Fl_Group*>(child);
		if (g)
		{
			::PrvClearDlgWidgetIDs (g);
		}
		else
		{
			::PrvClearWidgetID (child);
		}
	}
}

// Define data structures and routines for keeping track of
// strings that are associated with widgets.  We need to keep
// track of what strings have been assigned to which widgets
// because FLTK widgets don't *own* the labels assigned to
// them.  They expect the caller to own the storage.  However,
// that's not the model EmDlg uses; it assumes that the widgets
// own the storage (as is the case in PowerPlant and with
// Windows).
//
// So when we set the label of a widget, we keep ownership
// of the string by storing it in a private collection.  This
// collection is a "list" object.  I chose that one so that
// strings currently in the collection are undisturbed when
// new strings are added.  Widgets keep a pointer to the
// characters owned by the string object.  If that object
// moves around as the result of adding new string objects,
// it might invalidate the pointer owned by the widget.

typedef pair<Fl_Widget*, string>		WidgetString;
typedef list<WidgetString>				WidgetStringList;

WidgetStringList gWidgetStrings;

const char* PrvSetWidgetString (Fl_Widget* o, const string& s)
{
	// First see if there's already an entry for this widget.

	WidgetStringList::iterator iter = gWidgetStrings.end ();
	while (iter != gWidgetStrings.end ())
	{
		if (iter->first == o)
		{
			// Widget is already in the collection.
			// Replace the string it owns.

			iter->second = s;
			return iter->second.c_str ();
		}

		++iter;
	}

	// Widget is not in the collection, so add an entry for it.

	gWidgetStrings.push_back (WidgetString (o, s));
	return (--(gWidgetStrings.end()))->second.c_str ();
}

void PrvClearWidgetString (Fl_Widget* o)
{
	WidgetStringList::iterator iter = gWidgetStrings.end ();
	while (iter != gWidgetStrings.end ())
	{
		if (iter->first == o)
		{
			gWidgetStrings.erase (iter);
			return;
		}

		++iter;
	}
}

void PrvClearAllWidgetStrings (void)
{
	gWidgetStrings.clear ();
}

void PrvClearDlgWidgetStrings (Fl_Group* w)
{
	for (int ii = 0; ii < w->children (); ++ii)
	{
		Fl_Widget* child = w->child (ii);
		::PrvClearWidgetString (child);

		Fl_Group* g = dynamic_cast<Fl_Group*>(child);
		if (g)
		{
			::PrvClearDlgWidgetStrings (g);
		}
	}
}


queue<Fl_Widget*> gWidgetQueue;


Fl_Widget* PrvReadQueue (void)
{
	Fl_Widget* result = 0;
	if (!gWidgetQueue.empty ())
	{
		result = gWidgetQueue.front (); // Note, this is "top()" on Windows, maybe others...
		gWidgetQueue.pop ();
	}

	return result;
}

void PrvCallback (Fl_Widget* w, void*)
{
//	EmDlgItemID id = ::PrvFindIDByWidget (w);
//	printf ("PrvCallback: id = 0x%08lX\n", (long) id);
	
//	EmDlgContext& context = *(EmDlgContext*) c;
	gWidgetQueue.push (w);
}


void PrvInstallCallback (Fl_Widget* o, void* data)
{
	// Install callbacks only for our widgets.  We shouldn't
	// mess up any other callbacks -- they may do something
	// important (such as the callback installed into scrollbars
	// in an Fl_Scroll pane).

	if (::PrvFindIDByWidget (o) != kDlgItemNone)
	{
		o->callback (&::PrvCallback, data);
	}
}


void PrvInstallCallbacks (Fl_Group* w, EmDlgContext* context)
{
	// Set self.
	::PrvInstallCallback (w, context);

	// Set children.
	for (int ii = 0; ii < w->children (); ++ii)
	{
		Fl_Widget* child = w->child (ii);

		Fl_Group* g = dynamic_cast<Fl_Group*>(child);
		if (g)
		{
			::PrvInstallCallbacks (g, context);
		}
		else
		{
			::PrvInstallCallback (child, context);
		}
	}
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::HostRunGetFile
 *
 * DESCRIPTION:	Platform-specific routine for getting the name of a
 *				file from the user.  This file name is used to load
 *				the file.
 *
 * PARAMETERS:	typeList - types of files user is allowed to select.
 *				ref - selected file is returned here. Valid only if
 *					function result is kDlgItemOK.
 *
 * RETURNED:	ID of dialog item that dismissed the dialog.
 *
 ***********************************************************************/

EmDlgItemID EmDlg::HostRunGetFile (EmFileRef& result,
								   const string& prompt,
								   const EmDirRef& defaultPath,
								   const EmFileTypeList& typeList)
{
	FilterList filter = ::PrvConvertTypeList (typeList);

	FileChooser chooser (defaultPath.IsSpecified () ? defaultPath.GetFullPath ().c_str () : NULL,
						 filter.c_str (), FileChooser::SINGLE, prompt.c_str ());

	chooser.show ();

	while (chooser.visible ())
		Fl::wait ();

	long count = chooser.count ();
	if (count == 0)
		return kDlgItemCancel;

	// Get a EmFileRef to the given file

	result = EmFileRef (chooser.value(1));

	return kDlgItemOK;
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::HostRunGetFileList
 *
 * DESCRIPTION:	Platform-specific routine for getting the name of a
 *				file from the user.  This file name is used to load
 *				the file.
 *
 * PARAMETERS:	typeList - types of files user is allowed to select.
 *				ref - selected file is returned here. Valid only if
 *					function result is kDlgItemOK.
 *
 * RETURNED:	ID of dialog item that dismissed the dialog.
 *
 ***********************************************************************/

EmDlgItemID EmDlg::HostRunGetFileList (EmFileRefList& results,
								   const string& prompt,
								   const EmDirRef& defaultPath,
								   const EmFileTypeList& typeList)
{
	FilterList filter = ::PrvConvertTypeList (typeList);

	FileChooser chooser (defaultPath.IsSpecified () ? defaultPath.GetFullPath ().c_str () : NULL,
						 filter.c_str (), FileChooser::MULTI, prompt.c_str ());

	chooser.show ();

	while (chooser.visible ())
		Fl::wait ();

	long count = chooser.count ();
	if (count == 0)
		return kDlgItemCancel;

	for (long ii = 1; ii <= count; ++ii)
	{
		results.push_back (EmFileRef (chooser.value (ii)));
	}

	return kDlgItemOK;
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::HostRunPutFile
 *
 * DESCRIPTION:	Platform-specific routine for getting the name of a
 *				file from the user.  This file name is used to save
 *				the file.
 *
 * PARAMETERS:	defName - default name for the file to be saved.
 *				ref - selected file is returned here. Valid only if
 *					function result is kDlgItemOK.
 *
 * RETURNED:	ID of dialog item that dismissed the dialog.
 *
 ***********************************************************************/

EmDlgItemID EmDlg::HostRunPutFile (EmFileRef& result,
								   const string& prompt,
								   const EmDirRef& defaultPath,
								   const EmFileTypeList& typeList,
								   const string& defName)
{
	FilterList filter = ::PrvConvertTypeList (typeList);

	FileChooser chooser (defaultPath.IsSpecified () ? defaultPath.GetFullPath ().c_str () : NULL,
						 filter.c_str (), FileChooser::CREATE, prompt.c_str ());

	chooser.show ();

	while (chooser.visible ())
		Fl::wait ();

	long count = chooser.count ();
	if (count == 0)
		return kDlgItemCancel;

	// Get a EmFileRef to the given file

	result = EmFileRef (chooser.value(1));

	return kDlgItemOK;
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::HostRunGetDirectory
 *
 * DESCRIPTION:	Platform-specific routine for getting the name of a
 *				file from the user.  This file name is used to load
 *				the file.
 *
 * PARAMETERS:	typeList - types of files user is allowed to select.
 *				ref - selected file is returned here. Valid only if
 *					function result is kDlgItemOK.
 *
 * RETURNED:	ID of dialog item that dismissed the dialog.
 *
 ***********************************************************************/

EmDlgItemID EmDlg::HostRunGetDirectory (EmDirRef& result,
								   const string& prompt,
								   const EmDirRef& /*defaultPath*/)
{
	return kDlgItemCancel;
}



/***********************************************************************
 *
 * FUNCTION:	HostRunSessionSave
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

EmDlgItemID	EmDlg::HostRunSessionSave (	const string& appName,
										const string& docName,
										Bool quitting)
{
	string saveChanges (Platform::GetString (kStr_SaveBeforeClosing));

	ParamList paramList;
	paramList.push_back (string ("%AppName"));
	paramList.push_back (appName);
	paramList.push_back (string ("%DocName"));
	paramList.push_back (docName);

	string msg = Errors::ReplaceParameters (saveChanges, paramList);

	int result = fl_choice (msg.c_str(),
							Platform::GetString (kStr_Cancel).c_str(),
							Platform::GetString (kStr_Yes).c_str(),
							Platform::GetString (kStr_No).c_str() );

	if (result == 1) // yes
		return kDlgItemOK;
	else if (result == 2) // no
		return kDlgItemDontSave;

	// result == 0
	return kDlgItemCancel;
}


/***********************************************************************
 *
 * FUNCTION:	HostRunDialog
 *
 * DESCRIPTION:	Common routine that handles the creation of a dialog,
 *				initializes it (via the dialog handler), fetches events,
 *				handles events (via the dialog handler), and closes
 *				the dialog.
 *
 * PARAMETERS:	fn - the custom dialog handler
 *				userData - custom data passed back to the dialog handler.
 *				dlgID - ID of dialog to create.
 *
 * RETURNED:	ID of dialog item that dismissed the dialog.
 *
 ***********************************************************************/

EmDlgItemID EmDlg::HostRunDialog (EmDlgFn fn, void* userData, EmDlgID dlgID)
{
	Fl_Window* dlg = ::PrvMakeDialog (dlgID);
	assert (dlg);

	EmDlgContext		context;
	context.fFn			= fn;
//	context.fFnResult	= filled in after fFn is called;
	context.fDlg		= (EmDlgRef) dlg;
	context.fDlgID		= dlgID;
//	context.fPanelID	= filled in by subroutines;
//	context.fItemID		= filled in by subroutines;
//	context.fCommandID	= filled in by subroutines;
//	context.fNeedsIdle	= filled in by c'tor to false;
	context.fUserData	= userData;

	if (context.Init () == kDlgResultClose)
		return kDlgItemNone;

	::PrvInstallCallbacks (dlg, &context);

	dlg->show();

	for (;;)
	{
		Fl_Widget* o = ::PrvReadQueue ();
		if (o)
		{
			EmDlgItemID itemID  = ::PrvFindIDByWidget (o);
			if (itemID != kDlgItemNone)
				if (context.Event (itemID) == kDlgResultClose)
					break;
		}
		else
		{
			// Set this to see if the idle function changes it.
			context.fFnResult = kDlgResultNone;

			Fl::wait ();

			// See if the idle function wants to close the dialog.
			if (context.fFnResult == kDlgResultClose)
				break;
		}
	}

	HostStopIdling (context);

	dlg->hide ();

	::PrvClearDlgWidgetIDs (dlg);
	::PrvClearDlgWidgetStrings (dlg);

	return context.fItemID;
}


/***********************************************************************
 *
 * FUNCTION:	HostStartIdling
 *
 * DESCRIPTION:	Queue up our idle callback function with FLTK.
 *
 * PARAMETERS:	context - context data to be provided to the callback
 *					function.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void EmDlg::HostStartIdling (EmDlgContext& context)
{
	Fl::add_timeout (.01, &::PrvIdleCallback, &context);
}


/***********************************************************************
 *
 * FUNCTION:	HostStopIdling
 *
 * DESCRIPTION:	Remove our idle callback function from FLTK.
 *
 * PARAMETERS:	context - context data to be provided to the callback
 *					function.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void EmDlg::HostStopIdling (EmDlgContext& context)
{
	Fl::remove_timeout (&::PrvIdleCallback, &context);
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::SetItemMin
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::SetItemMin (EmDlgRef dlg, EmDlgItemID item, long minValue)
{
	Fl_Widget* o = ::PrvFindWidgetByID (item);
	if (!o)
		return;

	Fl_Valuator* valuator = dynamic_cast<Fl_Valuator*> (o);
	if (valuator)
		valuator->minimum (minValue);
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::SetItemValue
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::SetItemValue (EmDlgRef dlg, EmDlgItemID item, long value)
{
	Fl_Widget* o = ::PrvFindWidgetByID (item);
	if (!o)
		return;

	// ...............................................................
	// Includes: Fl_Button, Fl_Light_Button, Fl_Check_Button,
	// Fl_Radio_Light_Button, Fl_Round_Button, Fl_Radio_Round_Button,
	// Fl_Repeat_Button, Fl_Radio_Buttion, Fl_Toggle_Button
	// ...............................................................

	Fl_Button* button = dynamic_cast<Fl_Button*> (o);
	if (button)
		button->value (value);

	// ...............................................................
	// Includes: Fl_Choice, Fl_Menu_Bar, Fl_Menu_Button
	// ...............................................................

	Fl_Menu_* menu = dynamic_cast<Fl_Menu_*> (o);
	if (menu)
	{
		for (int i=0; i<value; i++)
			if (menu->menu()[i].flags & FL_MENU_DIVIDER)
				value--;

		menu->value (value);
	}

	// ...............................................................
	// Includes: Fl_Browser, Fl_Hold_Browser, Fl_Multi_Browser,
	// Fl_Select_Browser (but not Fl_Browser_)
	// ...............................................................

	Fl_Browser* browser = dynamic_cast<Fl_Browser*> (o);
	if (browser)
		browser->value (value);

	// ...............................................................
	// Includes: Fl_Adjuster, Fl_Counter, Fl_Dial, Fl_Roller,
	// Fl_Scrollber, Fl_Slider, Fl_Value_Input, Fl_Value_Slider,
	// Fl_Value_Output
	// ...............................................................

	Fl_Valuator* valuator = dynamic_cast<Fl_Valuator*> (o);
	if (valuator)
		valuator->value (value);

	// ...............................................................
	// Includes: Fl_Input_, Fl_Input, Fl_Float_Input, Fl_Int_Input,
	// Fl_Multiline_Input, Fl_Secret_Input, Fl_Output and
	// Fl_Multiline_Output
	// ...............................................................

	Fl_Input_* input = dynamic_cast<Fl_Input_*> (o);
	if (input)
	{
		char buffer[20];
		sprintf (buffer, "%ld", (long) value);
		EmDlg::SetItemText (dlg, item, buffer);
	}

	// ...............................................................
	// Includes: Fl_Box
	// ...............................................................

	Fl_Box* box = dynamic_cast<Fl_Box*> (o);
	if (box)
	{
		char	buffer[20];
		::FormatInteger (buffer, value);
		EmDlg::SetItemText (dlg, item, buffer);
	}
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::SetItemMax
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::SetItemMax (EmDlgRef dlg, EmDlgItemID item, long maxValue)
{
	Fl_Widget* o = ::PrvFindWidgetByID (item);
	if (!o)
		return;

	// ...............................................................
	// Includes: Fl_Adjuster, Fl_Counter, Fl_Dial, Fl_Roller,
	// Fl_Scrollber, Fl_Slider, Fl_Value_Input, Fl_Value_Slider,
	// Fl_Value_Output
	// ...............................................................

	Fl_Valuator* valuator = dynamic_cast<Fl_Valuator*> (o);
	if (valuator)
		valuator->maximum (maxValue);
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::SetItemText
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::SetItemText (EmDlgRef dlg, EmDlgItemID item, string str)
{
	Fl_Widget* o = ::PrvFindWidgetByID (item);
	if (!o)
		return;

	Fl_Input_* in = dynamic_cast<Fl_Input_*> (o);
	if (in)
	{
		in->value (str.c_str ());
	}
	else
	{
		// Setting the text of other items is tricky.  That would be
		// done by calling Fl_Widget::label.  However, that function
		// does not copy the text -- just the text pointer.  That
		// means the string object (or whatever is holding the storage)
		// must exist after this function is called.

		// Additionally, it appears that we have to hide and then show
		// the widget in order to get it to update properly.  Merely
		// calling redraw will not erase the old text in boxes like
		// the file counter in the "Import Database" dialog.

		const char* s = ::PrvSetWidgetString (o, str);

		if (o->visible_r ())
		{
			o->hide ();
			o->label (s);
			o->show ();
		}
		else
		{
			o->label (s);
		}
	}
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::GetItemValue
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

long EmDlg::GetItemValue (EmDlgRef dlg, EmDlgItemID item)
{
	Fl_Widget* o = ::PrvFindWidgetByID (item);
	if (!o)
		return 0;

	// ...............................................................
	// Includes: Fl_Button, Fl_Light_Button, Fl_Check_Button,
	// Fl_Radio_Light_Button, Fl_Round_Button, Fl_Radio_Round_Button,
	// Fl_Repeat_Button, Fl_Radio_Buttion, Fl_Toggle_Button
	// ...............................................................

	Fl_Button* button = dynamic_cast<Fl_Button*> (o);
	if (button)
		return button->value ();

	// ...............................................................
	// Includes: Fl_Choice, Fl_Menu_Bar, Fl_Menu_Button
	// ...............................................................

	Fl_Menu_* menu = dynamic_cast<Fl_Menu_*> (o);
	if (menu)
	{
		long	result = menu->value();
		long	add = 0;
		
		for (int i=0; i<result; i++)
			if (menu->menu()[i].flags & FL_MENU_DIVIDER)
				add++;
				
		return result + add;;
	}

	// ...............................................................
	// Includes: Fl_Browser (but not Fl_Browser_)
	// ...............................................................

	Fl_Browser* browser = dynamic_cast<Fl_Browser*> (o);
	if (browser)
		return browser->value ();

	// ...............................................................
	// Includes: Fl_Adjuster, Fl_Counter, Fl_Dial, Fl_Roller,
	// Fl_Scrollber, Fl_Slider, Fl_Value_Input, Fl_Value_Slider,
	// Fl_Value_Output
	// ...............................................................

	Fl_Valuator* valuator = dynamic_cast<Fl_Valuator*> (o);
	if (valuator)
		return (long) valuator->value ();		// !!! Really is a "double"!

	// ...............................................................
	// Includes: Fl_Input_, Fl_Input, Fl_Float_Input, Fl_Int_Input,
	// Fl_Multiline_Input, Fl_Secret_Input, Fl_Output and
	// Fl_Multiline_Output
	// ...............................................................

	Fl_Input_* input = dynamic_cast<Fl_Input_*> (o);
	if (input)
	{
		const char* text = input->value ();
		long value = 0;
		sscanf (text, "%ld", &value);
		return value;
	}

	return 0;
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::GetItemText
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

string EmDlg::GetItemText (EmDlgRef dlg, EmDlgItemID item)
{
	Fl_Widget* o = ::PrvFindWidgetByID (item);
	if (!o)
		return string ();

	Fl_Input_* in = dynamic_cast<Fl_Input_*> (o);
	if (in)
		return string (in->value ());

	return string (o->label ());
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::EnableItem
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::EnableItem (EmDlgRef dlg, EmDlgItemID item)
{
	Fl_Widget* o = ::PrvFindWidgetByID (item);
	if (!o)
		return;

	o->activate ();
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::DisableItem
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::DisableItem (EmDlgRef dlg, EmDlgItemID item)
{
	Fl_Widget* o = ::PrvFindWidgetByID (item);
	if (!o)
		return;

	o->deactivate ();
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::ClearMenu
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::ClearMenu (EmDlgRef dlg, EmDlgItemID item)
{
	Fl_Widget* o = ::PrvFindWidgetByID (item);
	if (!o)
		return;

	Fl_Menu_* menu = dynamic_cast<Fl_Menu_*> (o);
	assert (menu);

	menu->clear ();
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::AppendToMenu
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::AppendToMenu (EmDlgRef dlg, EmDlgItemID item, const StringList& strList)
{
	Fl_Widget* o = ::PrvFindWidgetByID (item);
	if (!o)
		return;

	Fl_Menu_* menu = dynamic_cast<Fl_Menu_*> (o);
	assert (menu);

	StringList::const_iterator iter = strList.begin ();
	while (iter != strList.end ())
	{
		if (iter->empty())
		{
			menu->mode(menu->size()-2, FL_MENU_DIVIDER);
		}
		else
		{
			menu->add (iter->c_str(), 0, NULL, NULL, 0);
		}
		++iter;
	}
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::SelectMenuItem
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::SelectMenuItem (EmDlgRef dlg, EmDlgItemID item, long menuItem)
{
	Fl_Widget* o = ::PrvFindWidgetByID (item);
	if (!o)
		return;

	Fl_Menu_* menu = dynamic_cast<Fl_Menu_*> (o);
	assert (menu);

	for (int i=0; i<menuItem; i++)
		if (menu->menu()[i].flags & FL_MENU_DIVIDER)
			menuItem--;
	
	menu->value (menuItem);
	menu->redraw ();
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::ClearList
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::ClearList (EmDlgRef dlg, EmDlgItemID item)
{
	Fl_Widget* o = ::PrvFindWidgetByID (item);
	if (!o)
		return;

	Fl_Browser* b = dynamic_cast<Fl_Browser*> (o);
	assert (b);

	b->clear ();
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::EnableMenuItem
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::EnableMenuItem (EmDlgRef dlg, EmDlgItemID item, long menuItem)
{
	Fl_Widget* o = ::PrvFindWidgetByID (item);
	if (!o)
		return;

	Fl_Menu_* menu = dynamic_cast<Fl_Menu_*> (o);
	assert (menu);

	for (int i=0; i<menuItem; i++)
		if (menu->menu()[i].flags & FL_MENU_DIVIDER)
			menuItem--;

	/* DOLATER - kja: This won't work properly if we try to change a divider */
	menu->mode (menuItem, 0);
	menu->redraw ();
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::DisableMenuItem
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::DisableMenuItem (EmDlgRef dlg, EmDlgItemID item, long menuItem)
{
	Fl_Widget* o = ::PrvFindWidgetByID (item);
	if (!o)
		return;

	Fl_Menu_* menu = dynamic_cast<Fl_Menu_*> (o);
	assert (menu);

	for (int i=0; i<menuItem; i++)
		if (menu->menu()[i].flags & FL_MENU_DIVIDER)
			menuItem--;

	/* DOLATER - kja: This won't work properly if we try to change a divider */
	menu->mode (menuItem, FL_MENU_INACTIVE);
	menu->redraw ();
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::AppendToList
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::AppendToList (EmDlgRef dlg, EmDlgItemID item, const StringList& strList)
{
	Fl_Widget* o = ::PrvFindWidgetByID (item);
	if (!o)
		return;

	Fl_Browser* b = dynamic_cast<Fl_Browser*> (o);
	assert (b);

	StringList::const_iterator iter = strList.begin ();
	while (iter != strList.end ())
	{
		b->add (iter->c_str (), NULL);
		++iter;
	}
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::SelectListItems
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::SelectListItems (EmDlgRef dlg, EmDlgItemID item, const EmDlgItemIndexList& itemList)
{
	Fl_Widget* o = ::PrvFindWidgetByID (item);
	if (!o)
		return;

	Fl_Browser* b = dynamic_cast<Fl_Browser*> (o);
	assert (b);

	EmDlgItemIndexList::const_iterator iter = itemList.begin ();
	while (iter != itemList.end ())
	{
		b->select (*iter + 1);
		++iter;
	}
}


/***********************************************************************
 *
 * FUNCTION:	EmDlg::GetSelectedItems
 *
 * DESCRIPTION:	
 *
 * PARAMETERS:	None
 *
 * RETURNED:	Nothing
 *
 ***********************************************************************/

void EmDlg::GetSelectedItems (EmDlgRef dlg, EmDlgItemID item, EmDlgItemIndexList& itemList)
{
	Fl_Widget* o = ::PrvFindWidgetByID (item);
	if (!o)
		return;

	Fl_Browser* b = dynamic_cast<Fl_Browser*> (o);
	assert (b);

	for (int ii = 1; ii <= b->size (); ++ii)
	{
		if (b->selected (ii))
		{
			itemList.push_back (ii - 1);
		}
	}
}


/***********************************************************************
 *
 * FUNCTION:	PrvConvertTypeList
 *
 * DESCRIPTION:	.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

FilterList PrvConvertTypeList (const EmFileTypeList& typeList)
{
	FilterList	filter;

	filter += "{";

	EmFileTypeList::const_iterator	iter = typeList.begin ();

	while (iter != typeList.end ())
	{
		switch (*iter)
		{
			case kFileTypeNone:
				break;

			case kFileTypeApplication:
				break;

			case kFileTypeROM:
				::PrvAddFilter (filter, "*.[Rr][Oo][Mm]");
				break;

			case kFileTypeSession:
				::PrvAddFilter (filter, "*.[Pp][Ss][Ff]");
				break;

			case kFileTypePreference:
				break;

			case kFileTypePalmApp:
				::PrvAddFilter (filter, "*.[Pp][Rr][Cc]");
				break;

			case kFileTypePalmDB:
				::PrvAddFilter (filter, "*.[Pp][Dd][Bb]");
				break;

			case kFileTypePalmQA:
				::PrvAddFilter (filter, "*.[Pp][Qq][Aa]");
				break;

			case kFileTypeText:
				::PrvAddFilter (filter, "*.[Tt][Xx][Tt]");
				break;

			case kFileTypePicture:
				::PrvAddFilter (filter, "*.[Pp][Pp][Mm]");
				break;

			case kFileTypeSkin:
				::PrvAddFilter (filter, "*.[Ss][Kk][Ii][Nn]");
				break;

			case kFileTypeProfile:
				break;

			case kFileTypePalmAll:
				::PrvAddFilter (filter, "*.[Pp][Rr][Cc]|*.[Pp][Dd][Bb]|*.[Pp][Qq][Aa]");
				break;

			case kFileTypeAll:
				::PrvAddFilter (filter, "*");
				break;

			case kFileTypeLast:
				assert (false);
				break;
		}

		++iter;
	}

	filter += "}";

	return filter;
}


/***********************************************************************
 *
 * FUNCTION:	PrvAddFilter
 *
 * DESCRIPTION:	.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void PrvAddFilter (FilterList& filter, const char* pattern)
{
	if (filter.size() != 1)
		filter += "|";

	filter += pattern;
}


/***********************************************************************
 *
 * FUNCTION:	PrvMakeDialog
 *
 * DESCRIPTION:	.
 *
 * PARAMETERS:	None.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

Fl_Window* PrvMakeDialog (EmDlgID id)
{
	Fl_Window* w = NULL;

	switch (id)
	{
		case kDlgNone:
			break;

		case kDlgAboutBox:
			break;

		case kDlgSessionNew:
			w = ::PrvMakeSessionNewDialog ();
			break;

		case kDlgSessionSave:
			break;

		case kDlgHordeNew:
			w = ::PrvMakeNewHordeDialog ();
			break;

		case kDlgDatabaseImport:
			w = ::PrvMakeDatabaseImportDialog ();
			break;

		case kDlgDatabaseExport:
			break;

		case kDlgROMTransferQuery:
			w = ::PrvMakeROMTransferQueryDialog ();
			break;

		case kDlgROMTransferProgress:
			w = ::PrvMakeROMTransferProgressDialog ();
			break;

		case kDlgGremlinControl:
			break;

		case kDlgEditPreferences:
			break;

		case kDlgEditLogging:
			w = ::PrvMakeEditLoggingOptionsDialog ();
			break;

		case kDlgEditDebugging:
			break;

		case kDlgEditSkins:
			w = ::PrvMakeEditSkinsDialog ();
			break;

		case kDlgEditCards:
			w = ::PrvMakeCardOptionsDialog ();

		case kDlgCommonDialog:
			break;

		case kDlgSaveBound:
			break;
	}

	return w;
}


/***********************************************************************
 *
 * FUNCTION:	PrvIdleCallback
 *
 * DESCRIPTION:	Function called by FLTK when our timeout function has
 *				timed out.  Here we call the custom dialog handler.  If
 *				it doesn't indicate that it's time to close the dialog,
 *				requeue the timeout function.
 *
 * PARAMETERS:	data - callback data.  We store the pointer to the
 *					current dialog context here.
 *
 * RETURNED:	Nothing.
 *
 ***********************************************************************/

void PrvIdleCallback (void* data)
{
	assert (data);
	EmDlgContext& context = *(EmDlgContext*) data;

	// Call the idle function.  The result is stored in context.fFnResult
	// as well as returned from the function.  If we need to continue,
	// the requeue the idle function.

	if (context.Idle () != kDlgResultClose && context.fNeedsIdle)
		EmDlg::HostStartIdling (context);
}
